searchNodes=[{"doc":"GenStage and Broadway producers for EventStoreDB Volley provides a GenStage producer Volley.LinearSubscription and a GenStage/Broadway producer Volley.PersistentSubscription . Both of these subscription producers can read a stream from beginning to end and then keep up-to-date as new events are published to the EventStoreDB. These producers can be used to build a reactive, event-driven, eventually consistent system suitable for Event Sourcing. In terms of Event Sourcing, these producers can be used to build process managers, sagas, and read models. Linear vs. persistent subscriptions The Volley.LinearSubscription producer is a simpler subscription model which uses Spear.read_stream/3 and Spear.subscribe/4 to read an EventStoreDB stream in order. These subscriptions are called linear within Volley because the strict ordering forces the stream revision to be described as a point on a one-dimensional number line. Volley.LinearSubscription is a client-side subscription: the client is responsible for storing its stream revision. Volley.PersistentSubscription s use the Persistent Subscription feature of EventStoreDB to store stream revisions and perform back-pressure on the EventStoreDB server-side. Persistent subscriptions do not have strict ordering guarantees, which allows features like competing consumers, batch processing, and message-parking (with a built-in dead letter approach). See the EventStoreDB documentation on persistent subscriptions and Spear.connect_to_persistent_subscription/5 for more information. Linear subscriptions have less resource impact on the EventStoreDB but are less flexible than persistent subscriptions. Linear subscriptions are subject to head-of-line blocking: failing to process an event must halt the subscription in order to keep ordering. Persistent subscriptions offer more complex subscription strategies and can avoid head-of-line blocking but handlers may be more complex or difficult to write as they need to account for events potentially arriving out of order. Systems do not need to be limited to only one kind of event listener: a mix of linear and persistent subscriptions may be wise.","ref":"Volley.html","title":"Volley","type":"module"},{"doc":"A subscription which guarantees ordering A linear subscription consumes an EventStoreDB stream in order, as if subscribed via Spear.subscribe/4 . Linear subscriptions are simpler than persistent subscriptions and can be used in cases where unordered processing is too complicated or undesirable. Back-pressure This producer primarily makes use of GenStage 's buffering capabilities to provide back-pressure. A plain subscription through Spear.subscribe/4 has no back-pressure. For very large streams, a subscriber process may become overwhelmed as the process mailbox fills up with events as fast as they can be read from the EventStoreDB. This producer has two modes: subscription mode, where the stream is subscribed with Spear.subscribe/4 and events are emitted as soon as available reading mode, in which events are emitted on-demand of the consumer This producer starts up in reading mode and emits events on-demand as long as there are more events to be read. Once the producer reaches the current end of the stream, it subscribes using Spear.subscribe/4 in the switch to subscription mode. Once the producer has caught up to the end of the stream, it will only receive newly appended events, and so may be less likely to become overwhelmed. Sustained bursts in appends to the stream may eventually overfill the GenStage buffer, though. Writing handlers for linear subscriptions Special care must be taken when writing a consumer for linear subscriptions. Consumers must implement head-of-line blocking in order to preserve correct ordering of events. A consumer must meet these four requirements only one consumer may subscribe to each producer the consumer must Process.link/1 itself to the producer process the GenStage.init/1 callback is a suitable place to perform this linking the consumer must only ever exert a demand of 0 or 1 the consumer must curate its stream position Let's build a basic event handler for a linear subscription with the GenStage basics defmodule MyHandler do use GenStage def start_link ( _ ) do GenStage . start_link ( __MODULE__ , :ok ) end @impl GenStage def init ( :ok ) do { :consumer , :ok , subscribe_to : [ MyProducer ] } end @impl GenStage def handle_events ( events , _from , state ) do IO . inspect ( events , label : &quot;events&quot; ) { :noreply , [ ] , state } end end This is a very minimal consumer that starts up, subscribes to MyProducer , and handles events by outputting them with IO.inspect/2 . We can start up this handler and the producer like so: linear_subscription_settings = [ name : MyProducer , connection : MySpearClient , stream_name : &quot;some_stream&quot; , .. ] [ { Volley.LinearSubscription , linear_subscription_settings } , MyHandler ] |&gt; Supervisor . start_link ( strategy : :one_for_one ) This consumer doesn't follow our rules though. If we start it up, we'll see that we're handling multiple events at once, and if we restart it we'll see it start from the beginning of the stream. Let's restrict the number of events it can process at once by tuning the :max_demand down to 1 : @impl GenStage def init ( :ok ) do { :consumer , :ok , subscribe_to : [ { MyProducer , max_demand : 1 } ] } end Now we're handling events one-by-one in our handler so we can match on a single event in GenStage.handle_events/3 def handle_events ( [ event ] , _from , state ) do IO . inspect ( event . metadata . stream_revision , label : &quot;handling event no.&quot; ) .. Now if we start up our handler, we'll see it churning through each event in order. Let's introduce a bit of failure into the handler though. Defensive programming is typically discouraged in OTP applications, so let's do a raise/1 in our handling code to simulate a situation like a failing bang ( ! ) function or bad match: def handle_events ( [ event ] , _from , state ) do revision = event . metadata . stream_revision IO . inspect ( revision , label : &quot;handling event no.&quot; ) if revision == 42 do raise &quot;aaaaah!&quot; end { :noreply , [ ] , state } end Now if we run our pipeline on a stream longer than 42 events, the handler will crash. Since producers and consumers are not linked by default in GenStage , the exit of the consumer will leave the producer running. This means that we will see output like handling event no . : 41 handling event no . : 42 21 : 03 : 07.107 [ error ] GenServer # PID &lt; 0 . 266 . 0 &gt; terminating ** (RuntimeError) aaaaah! handling event no . : 43 The default restart strategy of our consumer will start a new process which will subscribe to the producer and handle the next event. This means that event 42 is effectively skipped which breaks ordered processing. To remedy this we need to notify the producer that the processing for an event has failed by linking together the producer and consumer. @impl GenStage def init ( :ok ) do MyProducer |&gt; GenServer . whereis ( ) |&gt; Process . link ( ) { :consumer , :ok , subscribe_to : [ { MyProducer , max_demand : 1 } ] } end Now when the handler exits on event 42, it will also exit the producer. With the producer and consumer that we have so far, this will result in the consumer restarting processing from the beginning of the stream. Again, this breaks ordering. We need the consumer to curate its position in the stream in order to keep a consistent order. For this example, we'll use :ets to hold the stream position of our handler in memory. This is useful and easy to set up for an example, but :ets is an in-memory cache which will clear out when the service stops. Production storage for stream positions should be more persistent: e.g. a PostgreSQL row or an mnesia record. Outside of our supervision tree for the producer and consumer we'll create an :ets table: :ets . new ( :stream_positions , [ :set , :public , :named_table ] ) And now add a function to our handler so the producer can restore a stream position from this table def fetch_stream_position! do case :ets . lookup ( :stream_positions , __MODULE__ ) do [ { __MODULE__ , position } ] -&gt; position [ ] -&gt; :start end end And add that MFA to the producer's options: linear_subscription_settings = [ name : MyProducer , connection : MySpearClient , stream_name : &quot;some_stream&quot; , restore_stream_position! : { MyHandler , :fetch_stream_position! , [ ] } ] Now the producer will fetch the current stream position on start-up, so even if the processes crash and need to be restarted, the handler will keep a consistent position in the subscription. Finally we'll store the stream position in the consumer. This should only occur after the consumer has done any side-effects or processing prone to failure. Ideally, the stream position should be persisted in a transaction with any side-effects. @impl GenStage def handle_events ( [ event ] , _from , state ) do revision = event . metadata . stream_revision IO . inspect ( revision , label : &quot;handling event no.&quot; ) if revision == 42 do raise &quot;aaaaah!&quot; end :ets . insert ( :stream_positions , { __MODULE__ , revision } ) { :noreply , [ ] , state } end With this final change our consumer will read each event in the stream in order, reach event 42, raise, retry event 42, raise, and then the supervisor process will shut down. This is the essence of head-of-line blocking: once the pipeline reaches an event which it cannot process, the entire pipeline is halted. This is generally an undesirable behavior: a code-wise or manual change is usually needed to resolve the blockage. Persistent subscriptions (see Volley.PersistentSubscription ) offer much more flexibility around ordering, batching, and concurrency thanks to the asynchronous ack and nack workflow and the EventStoreDB's parking system. Altogether our handler looks like this: defmodule MyHandler do use GenStage def start_link ( _ ) do GenStage . start_link ( __MODULE__ , :ok ) end @impl GenStage def init ( :ok ) do MyProducer |&gt; GenServer . whereis ( ) |&gt; Process . link ( ) { :consumer , :ok , subscribe_to : [ { MyProducer , max_demand : 1 } ] } end @impl GenStage def handle_events ( [ event ] , _from , state ) do revision = event . metadata . stream_revision IO . inspect ( revision , label : &quot;handling event no.&quot; ) if revision == 42 do raise &quot;aaaaah!&quot; end :ets . insert ( :stream_positions , { __MODULE__ , revision } ) { :noreply , [ ] , state } end def fetch_stream_position! do case :ets . lookup ( :stream_positions , __MODULE__ ) do [ { __MODULE__ , position } ] -&gt; position [ ] -&gt; :start end end end Configuration :connection - (required) a Spear.Connection.t/0 to use for connecting to the EventStoreDB :stream_name - (required) the EventStoreDB stream to read :restore_stream_position! - (required) an MFA to invoke (with apply/3 ) to retrieve the stream position on start-up of the subscription. This function should read from the source to which the consumer is writing the stream position. A positive integer, a Spear.Event.t/0 , or the atoms :start or :end may be returned. :start starts the subscription at the first event in the stream while end immediately subscribes the producer to the end of the stream. :read_opts - (default: [] ) options to pass to Spear.read_stream/3 . The :max_count option may be worth tuning to achieve good performance: a stream of very small events may benefit from the batch-reading of a large max-count while a stream of very large events may be overwhelmed by a large max-count and need smaller read sizes. Remaining options are passed to GenStage.start_link/3 and the {:producer, state, opts} tuple in GenStage.init/1 .","ref":"Volley.LinearSubscription.html","title":"Volley.LinearSubscription","type":"module"},{"doc":"A GenStage/Broadway producer for persistent subscriptions Persistent subscriptions are a feature of EventStoreDB which offload the responsibilities of tracking processed events, back-pressure, and subscriber dispatch to the EventStoreDB server. This allows subscribers to more easily implement complex subscription strategies, such as allowing multiple subscribers across services without gossip, avoiding head-of-line blocking, and enabling concurrent and batch processing schemes and rate-limiting. Broadway features an acknowledgement interface which integrates well with the persistent subscription Spear.ack/3 and Spear.nack/4 system. Consumers intending to use this producer should prefer writing handlers with Broadway (over GenStage ) where possible. Configuration :broadway? - (default: false ) whether to emit events as Broadway.Message.t/0 messages or as Spear.Event.t/0 events. true should be set for this option if this producer is being used in a Broadway topology, false for use in a GenStage pipeline. When set as true , this producer will set the ack interface for each message to Volley.PersistentSubscription.Acknowledger with the proper connection details for sending acks and nacks. When false , the Spear.Event.metadata.subscription field will be replaced with a Volley.PersistentSubscription.Subscription.t/0 struct with any necessary connection details. :subscriptions - (default: [] ) a list of subscription configurations. Broadway does not currently allow more than one producer in a topology, however one may wish to subscribe to multiple persistent subscriptions, potentially across EventStoreDBs. Since back-pressure is controlled on the EventStoreDB side, a handler may specify multiple subscriptions in a single producer without any special considerations. The schema of each subscription is as follows :connection - (required) a Spear.Connection.t/0 process which can either be specified as a PID or any valid GenServer.name/0 :stream_name - (required) the EventStoreDB stream :group_name - (required) the EventStoreDB group name :opts - (default: [] ) options to pass to Spear.connect_to_persistent_subscription/5 . The main use of this options field is to configure the :buffer_size option (default: 1 ). The :buffer_size option controls the number of events allowed to be sent to this producer before any events are acknowledged. Remaining options are passed to GenStage.start_link/3 and the {:producer, state, opts} tuple in GenStage.init/1 . Examples defmodule MyHandler do use Broadway alias Broadway.Message def start_link ( _opts ) do subscription_opts = [ broadway? : true , subscriptions : [ [ stream_name : &quot;MyStream&quot; , group_name : inspect ( __MODULE__ ) , connection : MyApp.SpearClient , opts : [ # 10 events allowed in-flight at a time buffer_size : 10 ] ] ] ] Broadway . start_link ( __MODULE__ , name : __MODULE__ , producer : [ module : { Volley.PersistentSubscription , subscription_opts } ] , processors : [ default : [ concurrency : 2 ] ] , batchers : [ default : [ concurrency : 1 , batch_size : 5 ] ] ) end @impl Broadway def handle_message ( :default , % Message { } = message , _context ) do message |&gt; Message . put_batcher ( :default ) end @impl Broadway def handle_batch ( :default , messages , _batch_info , context ) do # do something batchy with messages... end end","ref":"Volley.PersistentSubscription.html","title":"Volley.PersistentSubscription","type":"module"},{"doc":"A Broadway acknowledger for persistent subscription messages Makes use of the batch capabilities of Spear.ack/3 and Spear.nack/4 to efficiently acknowledge messages. This module is only availble if the :broadway dependency is available.","ref":"Volley.PersistentSubscription.Acknowledger.html","title":"Volley.PersistentSubscription.Acknowledger","type":"module"},{"doc":"A structure which stores information about a subscription","ref":"Volley.PersistentSubscription.Subscription.html","title":"Volley.PersistentSubscription.Subscription","type":"module"},{"doc":"A subscription structure which stores connection details for any persistent subscription Acks and nacks may be sent with the :connection and :ref fields. Spear . ack ( subscription . connection , :subscription . ref , event_ids ) Spear . nack ( subscription . connection , :subscription . ref , event_ids , action : :retry ) If the Volley.PersistentSubscription producer is used in GenStage-mode (with broadway?: false ), each Spear.Event.t/0 will have one of these structs in its .metadata.subscription field which can be used for acks and nacks.","ref":"Volley.PersistentSubscription.Subscription.html#t:t/0","title":"Volley.PersistentSubscription.Subscription.t/0","type":"type"},{"doc":"Changelog All notable changes to this project will be documented in this file. The format is based on Keep a Changelog , and this project adheres to Semantic Versioning .","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"Added Initial implementations of persistent and linear subscriptions","ref":"changelog.html#0-1-0-2021-04-26","title":"Changelog - 0.1.0 - 2021-04-26","type":"extras"}]